import aerospike
from microservice_template_core.tools.logger import get_logger
from microservice_template_core.settings import AerospikeConfig
from aerospike import predexp as predexp
from aerospike import exception as exception
from prometheus_client import Summary
from aerospike_helpers.operations import list_operations as listops

logger = get_logger()


class AerospikeClient(object):
    # Prometheus Metrics
    AEROSPIKE_CONNECTIONS = Summary('aerospike_connections_latency_seconds', 'Time spent processing connect to aerospike')
    AEROSPIKE_READ = Summary('aerospike_read_latency_seconds', 'Time spent processing read from aerospike')
    AEROSPIKE_WRITE = Summary('aerospike_write_latency_seconds', 'Time spent processing write to aerospike')
    AEROSPIKE_DELETE = Summary('aerospike_delete_latency_seconds', 'Time spent processing delete from aerospike')
    AEROSPIKE_UPDATE = Summary('aerospike_update_latency_seconds', 'Time spent processing update in aerospike')
    AEROSPIKE_SCAN = Summary('aerospike_scan_latency_seconds', 'Time spent processing scan in aerospike')
    AEROSPIKE_QUERY = Summary('aerospike_query_latency_seconds', 'Time spent processing query in aerospike')
    AEROSPIKE_CREATE_INDEX = Summary('aerospike_create_index_latency_seconds', 'Time spent creating index in aerospike')
    AEROSPIKE_INSERT_MESSAGE_TO_LIST = Summary('aerospike_insert_message_to_list_latency_seconds', 'Time spent creating index in aerospike')

    def __init__(self, aerospike_set: str, bin_index: dict):
        self.client = self.client_aerospike()
        self.records_result = []
        self.create_index_string(aerospike_set=aerospike_set, bin_index=bin_index)

    @staticmethod
    @AEROSPIKE_CONNECTIONS.time()
    def client_aerospike():
        config = {
            'hosts': [(AerospikeConfig.AEROSPIKE_HOST, AerospikeConfig.AEROSPIKE_PORT)],
            'policies': {
                'write': {
                    'total_timeout': 3000,
                    'max_retries': 5,
                    'sleep_between_retries': 1000
                },
                'read': {
                    'total_timeout': 3000,
                    'max_retries': 5,
                    'sleep_between_retries': 1000
                }
            }
        }

        try:
            client = aerospike.client(config).connect()
            logger.info(msg=f"Connected to Aerospike - {config['hosts']}")
        except Exception as err:
            client = None
            logger.error(msg=f"failed to connect to the cluster with: {err} - {config['hosts']}")

        return client

    @AEROSPIKE_WRITE.time()
    def put_message(self, aerospike_set, aerospike_key, aerospike_message):
        key = (AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set, aerospike_key)
        try:
            self.client.put(key, aerospike_message)
        except Exception as e:
            logger.error(msg=f"error: {e}")

    @AEROSPIKE_INSERT_MESSAGE_TO_LIST.time()
    def insert_message_to_list(self, aerospike_set, aerospike_key, bin_name, bin_value):
        key = (AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set, aerospike_key)
        policy = {
            "write_flags": aerospike.LIST_WRITE_ADD_UNIQUE | aerospike.LIST_WRITE_NO_FAIL,
            "list_order": aerospike.LIST_UNORDERED,
        }
        try:
            self.client.operate(key, [listops.list_insert(bin_name=bin_name, value=bin_value, policy=policy, index=0)])
        except Exception as e:
            logger.error(msg=f"error: {e}")

    @AEROSPIKE_DELETE.time()
    def delete_message(self, aerospike_set, aerospike_key):
        key = (AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set, aerospike_key)
        try:
            self.client.remove(key)
        except Exception as e:
            logger.error(msg=f"error: {e}")

    @AEROSPIKE_UPDATE.time()
    def update_bin(self, aerospike_set, aerospike_key, bin_name, bin_value):
        key = (AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set, aerospike_key)
        try:
            self.client.append(key, bin_name, bin_value)
        except Exception as e:
            logger.error(msg=f"error: {e}")

    @AEROSPIKE_READ.time()
    def read_message(self, aerospike_set, aerospike_key):
        try:
            key = (AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set, aerospike_key)
            (key, metadata, record) = self.client.get(key)
        except Exception as err:
            return []

        return record

    @AEROSPIKE_SCAN.time()
    def scan_keys(self, aerospike_set):
        s = self.client.scan(AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set)
        records = []

        def callback(input_tuple):
            (_, _, record) = input_tuple
            records.append(record)
            return records

        s.foreach(callback)

        return records

    @AEROSPIKE_QUERY.time()
    def query_messages_predexps(self, aerospike_set, predexps: list):
        self.records_result = []

        q = self.client.query(AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set)
        q.predexp(predexps)
        q.foreach(self.callback)

        return self.records_result

    @AEROSPIKE_CREATE_INDEX.time()
    def create_index_string(self, aerospike_set, bin_index: dict):
        """
        :param aerospike_set: table name
        :param bin_index: example {"alert_id": "integer"}. Possible values: "integer", "string"
        :return: None
        """

        for bin_name, bin_type in bin_index.items():
            try:
                if bin_type == 'string':
                    self.client.index_string_create(AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set, bin_name, f'{aerospike_set}_{bin_name}_idx')
                elif bin_type == 'integer':
                    self.client.index_integer_create(AerospikeConfig.AEROSPIKE_NAMESPACE, aerospike_set, bin_name, f'{aerospike_set}_{bin_name}_idx')
            except exception.IndexFoundError:
                pass

    def callback(self, input_tuple):
        (key, meta, rec) = input_tuple
        self.records_result.append((rec))
        return self.records_result


# aerospike_client = AerospikeClient(aerospike_set='test_append_set', bin_index={'guid': 'string'})
#
# aerospike_client.append_message_to_list(aerospike_set='test_append_set', aerospike_key=10, bin_name='actions', bin_value={'valu7': 1, 'value_3': 2})
#
# print(aerospike_client.read_message(aerospike_set='test_append_set', aerospike_key=10))



from microservice_template_core.tools.logger import get_logger
from microservice_template_core.settings import ServiceConfig, KafkaConfig
from confluent_kafka import Consumer

logger = get_logger()


class KafkaConsumeMessages(object):
    def __init__(self, kafka_topic):
        self.kafka_topic = kafka_topic

    def start_consumer(self):
        """
        Start consumer
        """

        logger.debug(f"Initializing Kafka Confluent Consumer - {ServiceConfig.SERVICE_NAME}. Topic - {self.kafka_topic}")

        consumer = Consumer(
            {
                'bootstrap.servers': KafkaConfig.KAFKA_SERVERS,
                'group.id': ServiceConfig.SERVICE_NAME,
                'auto.offset.reset': 'latest',
                'allow.auto.create.topics': True,
                'session.timeout.ms': KafkaConfig.consumer_session_timeout_ms,
                'heartbeat.interval.ms': KafkaConfig.consumer_heartbeat_interval_ms,
                'offset.store.method': 'broker'
            }
        )

        consumer.subscribe([self.kafka_topic])

        return consumer



from prometheus_client import multiprocess, start_http_server, Gauge, Counter, Summary, Histogram, Info, Enum


class Prom:
    core_endpoint_get_health_count = Counter('my_start', 'Description of counter')



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from microservice_template_core.settings import DbConfig

engine = create_engine(f"mariadb+mariadbconnector://{DbConfig.DATABASE_USER}:{DbConfig.DATABASE_PSWD}@{DbConfig.DATABASE_SERVER}:{DbConfig.DATABASE_PORT}/{DbConfig.DATABASE_SCHEMA}")
Session = sessionmaker(bind=engine)

Base = declarative_base()



# from opentelemetry import trace
# from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
# from opentelemetry.instrumentation.flask import FlaskInstrumentor
# from opentelemetry.instrumentation.requests import RequestsInstrumentor
# from opentelemetry.sdk.resources import SERVICE_NAME, Resource
# from opentelemetry.sdk.trace import TracerProvider
# from opentelemetry.sdk.trace.export import BatchSpanProcessor
# from microservice_template_core.settings import TracerConfig, ServiceConfig
# tracer = None
#
#
# def get_tracer():
#     global tracer
#     if not tracer:
#         resource = Resource(attributes={
#             "service.name": ServiceConfig.SERVICE_NAME
#         })
#
#         trace.set_tracer_provider(
#             TracerProvider(resource=resource)
#         )
#
#         otlp_exporter = OTLPSpanExporter(
#             endpoint=f"{TracerConfig.AGENT_HOSTNAME}:{TracerConfig.AGENT_PORT}",
#             insecure=True
#         )
#
#         trace.get_tracer_provider().add_span_processor(BatchSpanProcessor(otlp_exporter))
#
#         tracer = trace.get_tracer(__name__)
#
#     return tracer



import logging
from microservice_template_core.settings import LoggerConfig, ServiceConfig
import logging_loki
from multiprocessing import Queue
import uuid
# from opentelemetry import trace

logger = None


class SpanFormatter(logging.Formatter):
    def format(self, record):
        # trace_id = trace.get_current_span().get_span_context().trace_id
        trace_id = uuid.uuid4()
        if trace_id == 0:
            record.trace_id = None
        else:
            record.trace_id = trace_id
        return super().format(record)


def get_logger():
    global logger
    if not logger:
        logger = logging.getLogger(ServiceConfig.SERVICE_NAME)
        logger.setLevel(LoggerConfig.LOG_LEVEL)

        loki_handler = logging_loki.LokiQueueHandler(
            Queue(-1),
            url=f"http://{LoggerConfig.LOKI_SERVER}:{LoggerConfig.LOKI_PORT}/loki/api/v1/push",
            tags={
                "service": ServiceConfig.SERVICE_NAME,
                "namespace": ServiceConfig.SERVICE_NAMESPACE,
                "pod": ServiceConfig.POD_NAME
            },
            version="1",
        )
        loki_handler.setFormatter(
            SpanFormatter(
                '%(levelname)s %(message)s trace_id=%(trace_id)s'
            )
        )

        console_handler = logging.StreamHandler()

        logger.addHandler(loki_handler)
        logger.addHandler(console_handler)

    return logger




import os


class ServiceConfig:
    SERVICE_NAME = os.getenv('SERVICE_NAME', 'microservice_template_core')
    SERVICE_PORT = os.getenv('SERVICE_PORT', 8081)
    SERVER_NAME = os.getenv('SERVER_NAME', "0.0.0.0")
    URL_PREFIX = os.getenv('URL_PREFIX', '/api/v1')
    SERVICE_NAMESPACE = os.getenv('SERVICE_NAMESPACE', 'dev')
    POD_NAME = os.getenv('POD_NAME', '')
    configuration = {}


class LoggerConfig:
    LOG_LEVEL = os.getenv('LOG_LEVEL', 'DEBUG')
    LOGGING_VERBOSE = os.getenv('LOGGING_VERBOSE', False)
    LOKI_SERVER = os.getenv('LOKI_SERVER', '127.0.0.1')
    LOKI_PORT = os.getenv('LOKI_PORT', 3100)


class TracerConfig:
    AGENT_HOSTNAME = os.getenv('AGENT_HOSTNAME', '127.0.0.1')
    AGENT_PORT = int(os.getenv('AGENT_PORT', 55680))


class FlaskConfig:
    FLASK_DEBUG = os.getenv('FLASK_DEBUG', False)
    FLASK_THREADED = os.getenv('FLASK_THREADED', False)

    FLASK_JWT = os.getenv('FLASK_JWT', True)
    JWT_SECRET_KEY = os.getenv('JWT_SECRET_KEY', 'khFw8H5hP3gQ9kKS')
    JWT_DECODE_ALGORITHMS = os.getenv('JWT_DECODE_ALGORITHMS', ['RS256'])
    JWT_IDENTITY_CLAIM = os.getenv('JWT_IDENTITY_CLAIM', 'sub')
    JWT_USER_CLAIMS = os.getenv('JWT_USER_CLAIMS', 'authorities')
    PROPAGATE_EXCEPTIONS = os.getenv('PROPAGATE_EXCEPTIONS', True)


class DbConfig:
    USE_DB = os.getenv('USE_DB', False)
    DATABASE_SERVER = os.getenv('DATABASE_SERVER', '127.0.0.1')
    DATABASE_PORT = os.getenv('DATABASE_PORT', 3306)
    DATABASE_USER = os.getenv('DATABASE_USER', 'microservice_template_core')
    DATABASE_PSWD = os.getenv('DATABASE_PSWD', 'microservice_template_core')
    DATABASE_SCHEMA = os.getenv('DATABASE_SCHEMA', 'harp')
    SQLALCHEMY_DATABASE_URI = f'mysql+pymysql://{DATABASE_USER}:{DATABASE_PSWD}@' \
                              f'{DATABASE_SERVER}:{DATABASE_PORT}/{DATABASE_SCHEMA}'
    SQLALCHEMY_TRACK_MODIFICATIONS = os.getenv('SQLALCHEMY_TRACK_MODIFICATIONS', False)
    SQLALCHEMY_POOL_RECYCLE = os.getenv('', 300)


class AerospikeConfig:
    AEROSPIKE_USER = os.getenv('AEROSPIKE_USER', 'admin')
    AEROSPIKE_PASS = os.getenv('AEROSPIKE_PASS', 'admin')
    AEROSPIKE_HOST = os.getenv('AEROSPIKE_HOST', '127.0.0.1')
    AEROSPIKE_PORT = int(os.getenv('AEROSPIKE_PORT', 3000))
    AEROSPIKE_NAMESPACE = os.getenv('AEROSPIKE_NAMESPACE', 'harpia')


class KafkaConfig:
    KAFKA_USER = os.getenv('KAFKA_USER', 'admin')
    KAFKA_PASS = os.getenv('KAFKA_PASS', 'admin')
    KAFKA_SERVERS = os.getenv('KAFKA_SERVERS', '127.0.0.1:9092')
    producer_message_send_max_retries = os.getenv('producer_message_send_max_retries', 6)
    producer_retry_backoff_ms = os.getenv('producer_retry_backoff_ms', 5000)
    producer_queue_buffering_max_ms = os.getenv('producer_queue_buffering_max_ms', 10000)
    producer_queue_buffering_max_messages = os.getenv('producer_queue_buffering_max_messages', 100000)
    producer_request_timeout_ms = os.getenv('producer_request_timeout_ms', 30000)
    consumer_session_timeout_ms = os.getenv('consumer_session_timeout_ms', 30000)
    consumer_heartbeat_interval_ms = os.getenv('consumer_heartbeat_interval_ms', 15000)
